@@ -1,8 +1,8 @@
 /*
  * gaim - Gadu-Gadu Protocol Plugin
- * $Id: gg.c,v 1.2 2001/10/04 05:45:17 warmenhoven Exp $
+ * $Id: gg.c,v 1.16 2001/12/08 09:48:52 warmenhoven Exp $
  *
- * Copyright (C) 2001, Arkadiusz Mikiewicz <misiek@pld.ORG.PL>
+ * Copyright (C) 2001 Arkadiusz Mikiewicz <misiek@pld.ORG.PL>
  * 
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -20,7 +20,9 @@
  *
  */
 
+#ifdef HAVE_CONFIG_H
 #include <config.h>
+#endif
 
 #include <netdb.h>
 #include <unistd.h>
@@ -40,6 +42,7 @@
 #endif
 #ifdef HAVE_ICONV
 #include <iconv.h>
+#include "iconv_string.h"
 #endif
 /* Library from EKG (Eksperymentalny Klient Gadu-Gadu) */
 #include "libgg.h"
@@ -59,7 +62,10 @@
 
 #define AGG_GENDER_NONE -1
 
-#define AGG_PUBDIR_FORM "/appsvc/fmpubquery2.asp"
+#define AGG_PUBDIR_USERLIST_EXPORT_FORM "/appsvc/fmcontactsput.asp"
+#define AGG_PUBDIR_USERLIST_IMPORT_FORM "/appsvc/fmcontactsget.asp"
+#define AGG_PUBDIR_SEARCH_FORM "/appsvc/fmpubquery2.asp"
+#define AGG_REGISTER_DATA_FORM "/appsvc/fmregister.asp"
 #define AGG_PUBDIR_MAX_ENTRIES 200
 
 #define AGG_STATUS_AVAIL              _("Available")
@@ -70,14 +76,27 @@
 #define AGG_STATUS_INVISIBLE_FRIENDS  _("Invisible for friends only")
 #define AGG_STATUS_NOT_AVAIL          _("Unavailable")
 
+#define AGG_HTTP_NONE			0
+#define AGG_HTTP_SEARCH			1
+#define AGG_HTTP_USERLIST_IMPORT	2
+#define AGG_HTTP_USERLIST_EXPORT	3
+#define AGG_HTTP_USERLIST_DELETE	4
+#define AGG_HTTP_PASSWORD_CHANGE	5
+
+#define UC_NORMAL 2
+
 struct agg_data {
 	struct gg_session *sess;
+	int own_status;
 };
 
-struct agg_search {
+struct agg_http {
 	struct gaim_connection *gc;
-	gchar *search_data;
+	gchar *request;
+	gchar *form;
+	gchar *host;
 	int inpa;
+	int type;
 };
 
 static char *agg_name()
@@ -88,34 +107,11 @@ static char *agg_name()
 static gchar *charset_convert(const gchar *locstr, char *encsrc, char *encdst)
 {
 #ifdef HAVE_ICONV
-	gchar *dststr;
-	size_t loclen, dstlen;
-	gchar *fsave, *tsave;
-	size_t count;
-	static iconv_t cd = (iconv_t)(-1);
-
-	if (cd == (iconv_t)(-1)) {
-		cd = iconv_open(encdst, encsrc);
-		if (cd == (iconv_t)(-1)) {
-			return g_strdup(locstr);
-		}
-	}
-
-	loclen = strlen(locstr);
-	/* we are ready for multibyte conversions */
-	dstlen = MB_LEN_MAX * loclen;
-	dststr = g_new0(gchar, dstlen + 1);
-	fsave = (gchar *)locstr;
-	tsave = dststr;
-	count = iconv(cd, &fsave, &loclen, &tsave, &dstlen);
-	if (count == -1) {
-		g_free(dststr);
-		return g_strdup(locstr);
-	}
-	return dststr;
-#else
-	return g_strdup(locstr);
+	gchar *result = NULL;
+	if (iconv_string(encdst, encsrc, locstr, locstr + strlen(locstr) + 1, &result, NULL) >= 0)
+		return result;
 #endif
+	return g_strdup(locstr);
 }
 
 static gboolean invalid_uin(char *uin)
@@ -177,7 +173,7 @@ static gchar *find_local_charset(void)
 	return gg_localenc;
 }
 
-static char *handle_errcode(int errcode, gboolean show)
+static char *handle_errcode(struct gaim_connection *gc, int errcode)
 {
 	static char msg[AGG_BUF_LEN];
 
@@ -205,74 +201,51 @@ static char *handle_errcode(int errcode, gboolean show)
 		break;
 	}
 
-	if (show)
-		do_error_dialog(msg, _("Gadu-Gadu Error"));
+	hide_login_progress(gc, msg);
 
 	return msg;
 }
 
-static gchar *encode_postdata(const gchar *data)
-{
-	gchar *p = NULL;
-	int i, j = 0;
-	for (i = 0; i < strlen(data); i++) {
-		/* locale insensitive, doesn't reflect RFC (1738 section 2.2, 1866 section 8.2.1) */
-		if ((data[i] >= 'a' && data[i] <= 'z')
-		    || (data[i] >= 'A' && data[i] <= 'Z')
-		    || (data[i] >= '0' && data[i] <= '9')
-		    || data[i] == '=' || data[i] == '&'
-		    || data[i] == '\n' || data[i] == '\r' || data[i] == '\t' || data[i] == '\014') {
-			p = g_realloc(p, j + 1);
-			p[j] = data[i];
-			j++;
-		} else {
-			p = g_realloc(p, j + 4);	/* remember, sprintf appends a '\0' */
-			sprintf(p + j, "%%%02x", (unsigned char)data[i]);
-			j += 3;
-		}
-	}
-	p = g_realloc(p, j + 1);
-	p[j] = '\0';
-
-	if (p && strlen(p))
-		return p;
-	else
-		return g_strdup(data);
-}
-
 static void agg_set_away(struct gaim_connection *gc, char *state, char *msg)
 {
 	struct agg_data *gd = (struct agg_data *)gc->proto_data;
+	int status = gd->own_status;
 
 	if (gc->away)
 		gc->away = NULL;
 
 	if (!g_strcasecmp(state, AGG_STATUS_AVAIL))
-		gg_change_status(gd->sess, GG_STATUS_AVAIL);
+		status = GG_STATUS_AVAIL;
 	else if (!g_strcasecmp(state, AGG_STATUS_AVAIL_FRIENDS))
-		gg_change_status(gd->sess, GG_STATUS_AVAIL | GG_STATUS_FRIENDS_MASK);
+		status = GG_STATUS_AVAIL | GG_STATUS_FRIENDS_MASK;
 	else if (!g_strcasecmp(state, AGG_STATUS_BUSY)) {
-		gg_change_status(gd->sess, GG_STATUS_BUSY);
+		status = GG_STATUS_BUSY;
 		gc->away = "";
 	} else if (!g_strcasecmp(state, AGG_STATUS_BUSY_FRIENDS)) {
-		gg_change_status(gd->sess, GG_STATUS_BUSY | GG_STATUS_FRIENDS_MASK);
+		status =  GG_STATUS_BUSY | GG_STATUS_FRIENDS_MASK;
 		gc->away = "";
 	} else if (!g_strcasecmp(state, AGG_STATUS_INVISIBLE)) {
-		gg_change_status(gd->sess, GG_STATUS_INVISIBLE);
+		status = GG_STATUS_INVISIBLE;
 		gc->away = "";
 	} else if (!g_strcasecmp(state, AGG_STATUS_INVISIBLE_FRIENDS)) {
-		gg_change_status(gd->sess, GG_STATUS_INVISIBLE | GG_STATUS_FRIENDS_MASK);
+		status = GG_STATUS_INVISIBLE | GG_STATUS_FRIENDS_MASK;
 		gc->away = "";
 	} else if (!g_strcasecmp(state, AGG_STATUS_NOT_AVAIL)) {
-		gg_change_status(gd->sess, GG_STATUS_NOT_AVAIL);
+		status = GG_STATUS_NOT_AVAIL;
 		gc->away = "";
 	} else if (!g_strcasecmp(state, GAIM_AWAY_CUSTOM)) {
 		if (msg) {
-			gg_change_status(gd->sess, GG_STATUS_BUSY);
+			status = GG_STATUS_BUSY;
 			gc->away = "";
 		} else
-			gg_change_status(gd->sess, GG_STATUS_AVAIL);
+			status = GG_STATUS_AVAIL;
+
+		if (gd->own_status & GG_STATUS_FRIENDS_MASK)
+			status |= GG_STATUS_FRIENDS_MASK;
 	}
+
+	gd->own_status = status;
+	gg_change_status(gd->sess, status);
 }
 
 static gchar *get_away_text(int uc)
@@ -299,7 +272,7 @@ static gchar *get_away_text(int uc)
 	}
 }
 
-static GList *agg_away_states()
+static GList *agg_away_states(struct gaim_connection *gc)
 {
 	GList *m = NULL;
 
@@ -367,6 +340,7 @@ static void main_callback(gpointer data, gint source, GaimInputCondition cond)
 
 	if (!(e = gg_watch_fd(gd->sess))) {
 		debug_printf("main_callback: gg_watch_fd failed - CRITICAL!\n");
+		hide_login_progress(gc, _("Unable to read socket"));
 		signoff(gc);
 		return;
 	}
@@ -381,7 +355,7 @@ static void main_callback(gpointer data, gint source, GaimInputCondition cond)
 	case GG_EVENT_CONN_FAILED:
 		if (gc->inpa)
 			gaim_input_remove(gc->inpa);
-		handle_errcode(e->event.failure, TRUE);
+		handle_errcode(gc, e->event.failure);
 		signoff(gc);
 		break;
 	case GG_EVENT_MSG:
@@ -389,11 +363,13 @@ static void main_callback(gpointer data, gint source, GaimInputCondition cond)
 			gchar *imsg;
 			gchar user[20];
 
-			g_snprintf(user, sizeof(user), "%u", e->event.msg.sender);
+			g_snprintf(user, sizeof(user), "%lu", e->event.msg.sender);
 			if (!allowed_uin(gc, user))
 				break;
 			imsg = charset_convert(e->event.msg.message, "CP1250", find_local_charset());
-			serv_got_im(gc, user, imsg, 0, time((time_t) NULL));
+			strip_linefeed(imsg);
+			/* e->event.msg.time - we don't know what this time is for */
+			serv_got_im(gc, user, imsg, 0, time(NULL), -1);
 			g_free(imsg);
 		}
 		break;
@@ -411,15 +387,14 @@ static void main_callback(gpointer data, gint source, GaimInputCondition cond)
 				case GG_STATUS_AVAIL:
 				case GG_STATUS_BUSY:
 				case GG_STATUS_INVISIBLE:
-				case GG_STATUS_FRIENDS_MASK:
-					status = UC_NORMAL | (e->event.status.status << 5);
+					status = UC_NORMAL | (n->status << 5);
 					break;
 				default:
 					status = UC_NORMAL;
 					break;
 				}
 
-				g_snprintf(user, sizeof(user), "%u", n->uin);
+				g_snprintf(user, sizeof(user), "%lu", n->uin);
 				serv_got_update(gc, user, (status == UC_UNAVAILABLE) ? 0 : 1, 0, 0, 0,
 						status, 0);
 				n++;
@@ -438,7 +413,6 @@ static void main_callback(gpointer data, gint source, GaimInputCondition cond)
 			case GG_STATUS_AVAIL:
 			case GG_STATUS_BUSY:
 			case GG_STATUS_INVISIBLE:
-			case GG_STATUS_FRIENDS_MASK:
 				status = UC_NORMAL | (e->event.status.status << 5);
 				break;
 			default:
@@ -446,9 +420,9 @@ static void main_callback(gpointer data, gint source, GaimInputCondition cond)
 				break;
 			}
 
-			g_snprintf(user, sizeof(user), "%u", e->event.status.uin);
-			serv_got_update(gc, user, (status == UC_UNAVAILABLE) ? 0 : 1, 0, 0, 0, status,
-					0);
+			g_snprintf(user, sizeof(user), "%lu", e->event.status.uin);
+			serv_got_update(gc, user, (status == UC_UNAVAILABLE) ? 0 : 1, 0, 0, 0,
+					status, 0);
 		}
 		break;
 	case GG_EVENT_ACK:
@@ -486,16 +460,17 @@ static void login_callback(gpointer data, gint source, GaimInputCondition cond)
 		gc->inpa = gaim_input_add(gd->sess->fd, GAIM_INPUT_READ, login_callback, gc);
 
 	switch (gd->sess->state) {
-	case GG_STATE_CONNECTING_HTTP:
-	case GG_STATE_WRITING_HTTP:
-		set_login_progress(gc, 2, _("Handshake"));
-	case GG_STATE_CONNECTING_GG:
-		set_login_progress(gc, 3, _("Connecting to GG server"));
+	case GG_STATE_READING_DATA:
+		set_login_progress(gc, 2, _("Reading data"));
 		break;
-	case GG_STATE_WAITING_FOR_KEY:
-		set_login_progress(gc, 4, _("Waiting for server key"));
-	case GG_STATE_SENDING_KEY:
-		set_login_progress(gc, 5, _("Sending key"));
+	case GG_STATE_CONNECTING_GG:
+		set_login_progress(gc, 3, _("Balancer handshake"));
+		break;
+	case GG_STATE_READING_KEY:
+		set_login_progress(gc, 4, _("Reading server key"));
+		break;
+	case GG_STATE_READING_REPLY:
+		set_login_progress(gc, 5, _("Exchanging key hash"));
 		break;
 	default:
 		break;
@@ -503,6 +478,7 @@ static void login_callback(gpointer data, gint source, GaimInputCondition cond)
 
 	if (!(e = gg_watch_fd(gd->sess))) {
 		debug_printf("login_callback: gg_watch_fd failed - CRITICAL!\n");
+		hide_login_progress(gc, _("Critical error in GG library\n"));
 		signoff(gc);
 		return;
 	}
@@ -521,12 +497,13 @@ static void login_callback(gpointer data, gint source, GaimInputCondition cond)
 		account_online(gc);
 		serv_finish_login(gc);
 
-		do_import(gc, NULL);
+		if (bud_list_cache_exists(gc))
+			do_import(gc, NULL);
 		break;
 	case GG_EVENT_CONN_FAILED:
 		gaim_input_remove(gc->inpa);
 		gc->inpa = 0;
-		handle_errcode(e->event.failure, TRUE);
+		handle_errcode(gc, e->event.failure);
 		signoff(gc);
 		break;
 	default:
@@ -540,6 +517,7 @@ static void agg_keepalive(struct gaim_connection *gc)
 {
 	struct agg_data *gd = (struct agg_data *)gc->proto_data;
 	if (gg_ping(gd->sess) < 0) {
+		hide_login_progress(gc, _("Unable to ping server"));
 		signoff(gc);
 		return;
 	}
@@ -551,6 +529,8 @@ static void agg_login(struct aim_user *user)
 	struct agg_data *gd = gc->proto_data = g_new0(struct agg_data, 1);
 	char buf[80];
 
+	gc->checkbox = _("Send as message");
+
 	gd->sess = g_new0(struct gg_session, 1);
 
 	if (user->proto_opt[USEROPT_NICK][0])
@@ -580,10 +560,9 @@ static void agg_login(struct aim_user *user)
 
 	gd->sess->uin = (uin_t) strtol(user->username, (char **)NULL, 10);
 	gd->sess->password = g_strdup(user->password);
-	gd->sess->state = GG_STATE_CONNECTING_HTTP;
+	gd->sess->state = GG_STATE_CONNECTING;
 	gd->sess->check = GG_CHECK_WRITE;
 	gd->sess->async = 1;
-	gd->sess->recv_left = 0;
 	gd->sess->fd = proxy_connect(GG_APPMSG_HOST, GG_APPMSG_PORT, login_callback, gc);
 
 	if (gd->sess->fd < 0) {
@@ -602,6 +581,7 @@ static void agg_close(struct gaim_connection *gc)
 	gg_logoff(gd->sess);
 	gg_free_session(gd->sess);
 	g_free(gc->proto_data);
+	gd->own_status = GG_STATUS_NOT_AVAIL;
 }
 
 static int agg_send_im(struct gaim_connection *gc, char *who, char *msg, int flags)
@@ -617,7 +597,8 @@ static int agg_send_im(struct gaim_connection *gc, char *who, char *msg, int fla
 
 	if (strlen(msg) > 0) {
 		imsg = charset_convert(msg, find_local_charset(), "CP1250");
-		if (gg_send_message(gd->sess, (flags & IM_FLAG_CHECKBOX) ? GG_CLASS_MSG : GG_CLASS_CHAT,
+		if (gg_send_message(gd->sess, (flags & IM_FLAG_CHECKBOX)
+				    ? GG_CLASS_MSG : GG_CLASS_CHAT,
 				    strtol(who, (char **)NULL, 10), imsg) < 0)
 			return -1;
 		g_free(imsg);
@@ -633,7 +614,7 @@ static void agg_add_buddy(struct gaim_connection *gc, char *who)
 	gg_add_notify(gd->sess, strtol(who, (char **)NULL, 10));
 }
 
-static void agg_rem_buddy(struct gaim_connection *gc, char *who)
+static void agg_rem_buddy(struct gaim_connection *gc, char *who, char *group)
 {
 	struct agg_data *gd = (struct agg_data *)gc->proto_data;
 	if (invalid_uin(who))
@@ -663,53 +644,15 @@ static void agg_add_buddies(struct gaim_connection *gc, GList *whos)
 	}
 }
 
-static void search_results(gpointer data, gint source, GaimInputCondition cond)
+static void search_results(struct gaim_connection *gc, gchar *webdata)
 {
-	struct agg_search *srch = data;
-	struct gaim_connection *gc = srch->gc;
+	gchar **webdata_tbl;
 	gchar *buf;
 	char *ptr;
-	char *webdata;
-	int len;
-	char read_data;
-	gchar **webdata_tbl;
 	int i, j;
 
-	if (!g_slist_find(connections, gc)) {
-		debug_printf("search_callback: g_slist_find error\n");
-		gaim_input_remove(srch->inpa);
-		g_free(srch);
-		close(source);
-		return;
-	}
-
-	webdata = NULL;
-	len = 0;
-
-	while (read(source, &read_data, 1) > 0 || errno == EWOULDBLOCK) {
-		if (errno == EWOULDBLOCK) {
-			errno = 0;
-			continue;
-		}
-
-		if (!read_data)
-			continue;
-
-		len++;
-		webdata = g_realloc(webdata, len);
-		webdata[len - 1] = read_data;
-	}
-
-	webdata = g_realloc(webdata, len + 1);
-	webdata[len] = 0;
-
-	gaim_input_remove(srch->inpa);
-	g_free(srch);
-	close(source);
-
 	if ((ptr = strstr(webdata, "query_results:")) == NULL || (ptr = strchr(ptr, '\n')) == NULL) {
 		debug_printf("search_callback: pubdir result [%s]\n", webdata);
-		g_free(webdata);
 		do_error_dialog(_("Couldn't get search results"), _("Gadu-Gadu Error"));
 		return;
 	}
@@ -719,8 +662,6 @@ static void search_results(gpointer data, gint source, GaimInputCondition cond)
 
 	webdata_tbl = g_strsplit(ptr, "\n", AGG_PUBDIR_MAX_ENTRIES);
 
-	g_free(webdata);
-
 	j = 0;
 
 	/* Parse array */
@@ -805,101 +746,424 @@ static void search_results(gpointer data, gint source, GaimInputCondition cond)
 
 	g_strfreev(webdata_tbl);
 
-	g_show_info_text(buf, NULL);
+	g_show_info_text(gc, NULL, 2, buf, NULL);
 
 	g_free(buf);
 }
 
-static void search_callback(gpointer data, gint source, GaimInputCondition cond)
+static void import_buddies_server_results(struct gaim_connection *gc, gchar *webdata)
 {
-	struct agg_search *srch = data;
-	struct gaim_connection *gc = srch->gc;
-	gchar *search_data = srch->search_data;
-	gchar *buf;
-	char *ptr;
+	gchar *ptr;
+	gchar **users_tbl;
+	int i;
+	if (strstr(webdata, "no_data:")) {
+		do_error_dialog(_("There is no Buddy List stored on server. Sorry!"),
+				_("Gadu-Gadu Error"));
+		return;
+	}
 
-	debug_printf("search_callback enter: begin\n");
+	if ((ptr = strstr(webdata, "get_results:")) == NULL || (ptr = strchr(ptr, ':')) == NULL) {
+		debug_printf("import_buddies_server_results: import buddies result [%s]\n", webdata);
+		do_error_dialog(_("Couldn't Import Buddies List from Server"), _("Gadu-Gadu Error"));
+		return;
+	}
+	ptr++;
+
+	users_tbl = g_strsplit(ptr, "\n", AGG_PUBDIR_MAX_ENTRIES);
+
+	/* Parse array of Buddies List */
+	for (i = 0; users_tbl[i] != NULL; i++) {
+		gchar **data_tbl;
+		gchar *name, *show;
+
+		g_strdelimit(users_tbl[i], "\r\t\n\015", ' ');
+		data_tbl = g_strsplit(users_tbl[i], ";", 8);
+
+		show = data_tbl[3];
+		name = data_tbl[6];
+
+		if (invalid_uin(name)) {
+			continue;
+		}
+
+		debug_printf("import_buddies_server_results: uin: %s\n", name);
+		if (!find_buddy(gc, name)) {
+			/* Default group if none specified on server */
+			gchar *group = g_strdup("Gadu-Gadu");
+			if (strlen(data_tbl[5])) {
+				gchar **group_tbl = g_strsplit(data_tbl[5], ",", 2);
+				if (strlen(group_tbl[0])) {
+					g_free(group);
+					group = g_strdup(group_tbl[0]);
+				}
+				g_strfreev(group_tbl);
+			}
+			/* Add Buddy to our userlist */
+			add_buddy(gc, group, name, strlen(show) ? show : name);
+			do_export(gc);
+			g_free(group);
+		}
+		g_strfreev(data_tbl);
+	}
+	g_strfreev(users_tbl);
+}
+
+static void export_buddies_server_results(struct gaim_connection *gc, gchar *webdata)
+{
+	if (strstr(webdata, "put_success:")) {
+		do_error_dialog(_("Buddies List sucessfully transfered into Server"),
+				_("Gadu-Gadu Information"));
+		return;
+	}
+
+	debug_printf("export_buddies_server_results: webdata [%s]\n", webdata);
+	do_error_dialog(_("Couldn't transfer Buddies List into Server"), _("Gadu-Gadu Error"));
+}
+
+static void delete_buddies_server_results(struct gaim_connection *gc, gchar *webdata)
+{
+	if (strstr(webdata, "put_success:")) {
+		do_error_dialog(_("Buddies List sucessfully deleted from Server"),
+				_("Gadu-Gadu Information"));
+		return;
+	}
+
+	debug_printf("delete_buddies_server_results: webdata [%s]\n", webdata);
+	do_error_dialog(_("Couldn't delete Buddies List from Server"), _("Gadu-Gadu Error"));
+}
+
+static void password_change_server_results(struct gaim_connection *gc, gchar *webdata)
+{
+	if (strstr(webdata, "reg_success:")) {
+		do_error_dialog(_("Password changed sucessfully"),
+				_("Gadu-Gadu Information"));
+		return;
+	}
+
+	debug_printf("delete_buddies_server_results: webdata [%s]\n", webdata);
+	do_error_dialog(_("Password couldn't be changed"), _("Gadu-Gadu Error"));
+}
+
+static void http_results(gpointer data, gint source, GaimInputCondition cond)
+{
+	struct agg_http *hdata = data;
+	struct gaim_connection *gc = hdata->gc;
+	char *webdata;
+	int len;
+	char read_data;
+
+	debug_printf("http_results: begin\n");
 
 	if (!g_slist_find(connections, gc)) {
 		debug_printf("search_callback: g_slist_find error\n");
-		g_free(search_data);
-		g_free(srch);
+		gaim_input_remove(hdata->inpa);
+		g_free(hdata);
+		close(source);
+		return;
+	}
+
+	webdata = NULL;
+	len = 0;
+
+	while (read(source, &read_data, 1) > 0 || errno == EWOULDBLOCK) {
+		if (errno == EWOULDBLOCK) {
+			errno = 0;
+			continue;
+		}
+
+		if (!read_data)
+			continue;
+
+		len++;
+		webdata = g_realloc(webdata, len);
+		webdata[len - 1] = read_data;
+	}
+
+	webdata = g_realloc(webdata, len + 1);
+	webdata[len] = 0;
+
+	gaim_input_remove(hdata->inpa);
+	close(source);
+
+	debug_printf("http_results: type %d, webdata [%s]\n", hdata->type, webdata);
+
+	switch (hdata->type) {
+	case AGG_HTTP_SEARCH:
+		search_results(gc, webdata);
+		break;
+	case AGG_HTTP_USERLIST_IMPORT:
+		import_buddies_server_results(gc, webdata);
+		break;
+	case AGG_HTTP_USERLIST_EXPORT:
+		export_buddies_server_results(gc, webdata);
+		break;
+	case AGG_HTTP_USERLIST_DELETE:
+	        delete_buddies_server_results(gc, webdata);
+		break;
+	case AGG_HTTP_PASSWORD_CHANGE:
+		password_change_server_results(gc, webdata);
+		break;
+	case AGG_HTTP_NONE:
+	default:
+		debug_printf("http_results: unsupported type %d\n", hdata->type);
+		break;
+	}
+
+	g_free(webdata);
+	g_free(hdata);
+}
+
+static void http_req_callback(gpointer data, gint source, GaimInputCondition cond)
+{
+	struct agg_http *hdata = data;
+	struct gaim_connection *gc = hdata->gc;
+	gchar *request = hdata->request;
+	gchar *buf;
+
+	debug_printf("http_req_callback: begin\n");
+
+	if (!g_slist_find(connections, gc)) {
+		debug_printf("http_req_callback: g_slist_find error\n");
+		g_free(request);
+		g_free(hdata);
 		close(source);
 		return;
 	}
 
 	if (source == -1) {
-		g_free(search_data);
-		g_free(srch);
+		g_free(request);
+		g_free(hdata);
 		return;
 	}
 
-	ptr = encode_postdata(search_data);
-	g_free(search_data);
+	debug_printf("http_req_callback: http request [%s]\n", request);
 
-	debug_printf("search_callback: pubdir request [%s]\n", ptr);
-
-	buf = g_strdup_printf("POST " AGG_PUBDIR_FORM " HTTP/1.0\r\n"
-			      "Host: " GG_PUBDIR_HOST "\r\n"
+	buf = g_strdup_printf("POST %s HTTP/1.0\r\n"
+			      "Host: %s\r\n"
 			      "Content-Type: application/x-www-form-urlencoded\r\n"
-			      "User-Agent: Mozilla/4.7 [en] (Win98; I)\r\n"
+			      "User-Agent: " GG_HTTP_USERAGENT "\r\n"
 			      "Content-Length: %d\r\n"
-			      "Pragma: no-cache\r\n" "\r\n" "%s\r\n", strlen(ptr), ptr);
+			      "Pragma: no-cache\r\n" "\r\n" "%s\r\n",
+			      hdata->form, hdata->host, strlen(request), request);
 
-	g_free(ptr);
+	g_free(request);
 
 	if (write(source, buf, strlen(buf)) < strlen(buf)) {
 		g_free(buf);
-		g_free(srch);
+		g_free(hdata);
 		close(source);
-		do_error_dialog(_("Couldn't send search request"), _("Gadu-Gadu Error"));
+		do_error_dialog(_("Couldn't send http request"), _("Gadu-Gadu Error"));
 		return;
 	}
 
 	g_free(buf);
 
-	srch->inpa = gaim_input_add(source, GAIM_INPUT_READ, search_results, srch);
+	hdata->inpa = gaim_input_add(source, GAIM_INPUT_READ, http_results, hdata);
+}
+
+static void import_buddies_server(struct gaim_connection *gc)
+{
+	struct agg_http *hi = g_new0(struct agg_http, 1);
+	static char msg[AGG_BUF_LEN];
+	gchar *u = gg_urlencode(gc->username);
+	gchar *p = gg_urlencode(gc->password);
+
+	hi->gc = gc;
+	hi->type = AGG_HTTP_USERLIST_IMPORT;
+	hi->form = AGG_PUBDIR_USERLIST_IMPORT_FORM;
+	hi->host = GG_PUBDIR_HOST;
+	hi->request = g_strdup_printf("FmNum=%s&Pass=%s", u, p);
+
+	g_free(u);
+	g_free(p);
+
+	if (proxy_connect(GG_PUBDIR_HOST, GG_PUBDIR_PORT, http_req_callback, hi) < 0) {
+		g_snprintf(msg, sizeof(msg), _("Buddies List import from Server failed (%s)"),
+			   GG_PUBDIR_HOST);
+		do_error_dialog(msg, _("Gadu-Gadu Error"));
+		g_free(hi->request);
+		g_free(hi);
+		return;
+	}
+}
+
+static void export_buddies_server(struct gaim_connection *gc)
+{
+	struct agg_http *he = g_new0(struct agg_http, 1);
+	static char msg[AGG_BUF_LEN];
+	gchar *ptr;
+	gchar *u = gg_urlencode(gc->username);
+	gchar *p = gg_urlencode(gc->password);
+
+	GSList *gr = gc->groups;
+
+	he->gc = gc;
+	he->type = AGG_HTTP_USERLIST_EXPORT;
+	he->form = AGG_PUBDIR_USERLIST_EXPORT_FORM;
+	he->host = GG_PUBDIR_HOST;
+	he->request = g_strdup_printf("FmNum=%s&Pass=%s&Contacts=", u, p);
+
+	g_free(u);
+	g_free(p);
+
+	while (gr) {
+		struct group *g = gr->data;
+		GSList *m = g->members;
+		while (m) {
+			struct buddy *b = m->data;
+			gchar *newdata;
+			/* GG Number */
+			gchar *name = gg_urlencode(b->name);
+			/* GG Pseudo */
+			gchar *show = gg_urlencode(strlen(b->show) ? b->show : b->name);
+			/* Group Name */
+			gchar *gname = gg_urlencode(g->name);
+
+			ptr = he->request;
+			newdata = g_strdup_printf("%s;%s;%s;%s;%s;%s;%s\r\n",
+						  show, show, show, show, "", gname, name);
+			he->request = g_strconcat(ptr, newdata, NULL);
+
+			g_free(newdata);
+			g_free(ptr);
+
+			g_free(gname);
+			g_free(show);
+			g_free(name);
+
+			m = g_slist_next(m);
+		}
+		gr = g_slist_next(gr);
+	}
+
+	if (proxy_connect(GG_PUBDIR_HOST, GG_PUBDIR_PORT, http_req_callback, he) < 0) {
+		g_snprintf(msg, sizeof(msg), _("Buddies List export to Server failed (%s)"),
+			   GG_PUBDIR_HOST);
+		do_error_dialog(msg, _("Gadu-Gadu Error"));
+		g_free(he->request);
+		g_free(he);
+		return;
+	}
+}
+
+static void delete_buddies_server(struct gaim_connection *gc)
+{
+	struct agg_http *he = g_new0(struct agg_http, 1);
+	static char msg[AGG_BUF_LEN];
+	gchar *u = gg_urlencode(gc->username);
+	gchar *p = gg_urlencode(gc->password);
+
+	he->gc = gc;
+	he->type = AGG_HTTP_USERLIST_DELETE;
+	he->form = AGG_PUBDIR_USERLIST_EXPORT_FORM;
+	he->host = GG_PUBDIR_HOST;
+	he->request = g_strdup_printf("FmNum=%s&Pass=%s&Delete=1", u, p);
+
+	if (proxy_connect(GG_PUBDIR_HOST, GG_PUBDIR_PORT, http_req_callback, he) < 0) {
+		g_snprintf(msg, sizeof(msg), _("Deletion of Buddies List from Server failed (%s)"),
+			   GG_PUBDIR_HOST);
+		do_error_dialog(msg, _("Gadu-Gadu Error"));
+		g_free(he->request);
+		g_free(he);
+		return;
+	}
 }
 
 static void agg_dir_search(struct gaim_connection *gc, char *first, char *middle,
-			   char *last, char *maiden, char *city, char *state, char *country, char *email)
+			   char *last, char *maiden, char *city, char *state,
+			   char *country, char *email)
 {
-	struct agg_search *srch = g_new0(struct agg_search, 1);
+	struct agg_http *srch = g_new0(struct agg_http, 1);
 	static char msg[AGG_BUF_LEN];
 
 	srch->gc = gc;
+	srch->type = AGG_HTTP_SEARCH;
+	srch->form = AGG_PUBDIR_SEARCH_FORM;
+	srch->host = GG_PUBDIR_HOST;
 
 	if (email && strlen(email)) {
-		srch->search_data = g_strdup_printf("Mode=1&Email=%s", email);
+		gchar *eemail = gg_urlencode(email);
+		srch->request = g_strdup_printf("Mode=1&Email=%s", eemail);
+		g_free(eemail);
 	} else {
 		gchar *new_first = charset_convert(first, find_local_charset(), "CP1250");
 		gchar *new_last = charset_convert(last, find_local_charset(), "CP1250");
 		gchar *new_city = charset_convert(city, find_local_charset(), "CP1250");
 
-		/* For active only add &ActiveOnly= */
-		srch->search_data = g_strdup_printf("Mode=0&FirstName=%s&LastName=%s&Gender=%d"
-						    "&NickName=%s&City=%s&MinBirth=%d&MaxBirth=%d",
-						    new_first, new_last, AGG_GENDER_NONE,
-						    "", new_city, 0, 0);
+		gchar *enew_first = gg_urlencode(new_first);
+		gchar *enew_last = gg_urlencode(new_last);
+		gchar *enew_city = gg_urlencode(new_city);
 
 		g_free(new_first);
 		g_free(new_last);
 		g_free(new_city);
+
+		/* For active only add &ActiveOnly= */
+		srch->request = g_strdup_printf("Mode=0&FirstName=%s&LastName=%s&Gender=%d"
+						"&NickName=%s&City=%s&MinBirth=%d&MaxBirth=%d",
+						enew_first, enew_last, AGG_GENDER_NONE,
+						"", enew_city, 0, 0);
+
+		g_free(enew_first);
+		g_free(enew_last);
+		g_free(enew_city);
 	}
 
-	if (proxy_connect(GG_PUBDIR_HOST, GG_PUBDIR_PORT, search_callback, srch) < 0) {
-		g_snprintf(msg, sizeof(msg), _("Connect to search service failed (%s)"), GG_PUBDIR_HOST);
+	if (proxy_connect(GG_PUBDIR_HOST, GG_PUBDIR_PORT, http_req_callback, srch) < 0) {
+		g_snprintf(msg, sizeof(msg), _("Connect to search service failed (%s)"),
+			   GG_PUBDIR_HOST);
 		do_error_dialog(msg, _("Gadu-Gadu Error"));
-		g_free(srch->search_data);
+		g_free(srch->request);
 		g_free(srch);
 		return;
 	}
 }
 
+static void agg_change_passwd(struct gaim_connection *gc, char *old, char *new)
+{
+	struct agg_http *hpass = g_new0(struct agg_http, 1);
+	static char msg[AGG_BUF_LEN];
+	gchar *u = gg_urlencode(gc->username);
+	gchar *p = gg_urlencode(gc->password);
+	gchar *enew = gg_urlencode(new);
+	gchar *eold = gg_urlencode(old);
+
+	hpass->gc = gc;
+	hpass->type = AGG_HTTP_PASSWORD_CHANGE;
+	hpass->form = AGG_REGISTER_DATA_FORM;
+	hpass->host = GG_REGISTER_HOST;
+
+	/* We are using old password as place for email - it's ugly */
+	hpass->request = g_strdup_printf("fmnumber=%s&fmpwd=%s&pwd=%s&email=%s&code=%u",
+					 u, p, enew, eold, gg_http_hash(old, new));
+
+	g_free(u);
+	g_free(p);
+	g_free(enew);
+	g_free(eold);
+
+	if (proxy_connect(GG_REGISTER_HOST, GG_REGISTER_PORT, http_req_callback, hpass) < 0) {
+		g_snprintf(msg, sizeof(msg), _("Changing Password failed (%s)"),
+			   GG_REGISTER_HOST);
+		do_error_dialog(msg, _("Gadu-Gadu Error"));
+		g_free(hpass->request);
+		g_free(hpass);
+		return;
+	}                                        
+}
+
 static void agg_do_action(struct gaim_connection *gc, char *action)
 {
 	if (!strcmp(action, _("Directory Search"))) {
 		show_find_info(gc);
+	} else if (!strcmp(action, _("Change Password"))) {
+		show_change_passwd(gc);
+	} else if (!strcmp(action, _("Import Buddies List from Server"))) {
+		import_buddies_server(gc);
+	} else if (!strcmp(action, _("Export Buddies List to Server"))) {
+		export_buddies_server(gc);
+	} else if (!strcmp(action, _("Delete Buddies List from Server"))) {
+	        delete_buddies_server(gc);
 	}
 }
 
@@ -908,33 +1172,46 @@ static GList *agg_actions()
 	GList *m = NULL;
 
 	m = g_list_append(m, _("Directory Search"));
+	m = g_list_append(m, NULL);
+	m = g_list_append(m, _("Change Password"));
+	m = g_list_append(m, NULL);
+	m = g_list_append(m, _("Import Buddies List from Server"));
+	m = g_list_append(m, _("Export Buddies List to Server"));
+	m = g_list_append(m, _("Delete Buddies List from Server"));
 
 	return m;
 }
 
 static void agg_get_info(struct gaim_connection *gc, char *who)
 {
-	struct agg_search *srch = g_new0(struct agg_search, 1);
+	struct agg_http *srch = g_new0(struct agg_http, 1);
 	static char msg[AGG_BUF_LEN];
 
 	srch->gc = gc;
+	srch->type = AGG_HTTP_SEARCH;
+	srch->form = AGG_PUBDIR_SEARCH_FORM;
+	srch->host = GG_PUBDIR_HOST;
 
 	/* If it's invalid uin then maybe it's nickname? */
 	if (invalid_uin(who)) {
 		gchar *new_who = charset_convert(who, find_local_charset(), "CP1250");
-
-		srch->search_data = g_strdup_printf("Mode=0&FirstName=%s&LastName=%s&Gender=%d"
-						    "&NickName=%s&City=%s&MinBirth=%d&MaxBirth=%d",
-						    "", "", AGG_GENDER_NONE, new_who, "", 0, 0);
-
+		gchar *enew_who = gg_urlencode(new_who);
+		
 		g_free(new_who);
-	} else
-		srch->search_data = g_strdup_printf("Mode=3&UserId=%s", who);
 
-	if (proxy_connect(GG_PUBDIR_HOST, GG_PUBDIR_PORT, search_callback, srch) < 0) {
-		g_snprintf(msg, sizeof(msg), _("Connect to search service failed (%s)"), GG_PUBDIR_HOST);
+		srch->request = g_strdup_printf("Mode=0&FirstName=%s&LastName=%s&Gender=%d"
+						"&NickName=%s&City=%s&MinBirth=%d&MaxBirth=%d",
+						"", "", AGG_GENDER_NONE, enew_who, "", 0, 0);
+
+		g_free(enew_who);
+	} else
+		srch->request = g_strdup_printf("Mode=3&UserId=%s", who);
+
+	if (proxy_connect(GG_PUBDIR_HOST, GG_PUBDIR_PORT, http_req_callback, srch) < 0) {
+		g_snprintf(msg, sizeof(msg), _("Connect to search service failed (%s)"),
+			   GG_PUBDIR_HOST);
 		do_error_dialog(msg, _("Gadu-Gadu Error"));
-		g_free(srch->search_data);
+		g_free(srch->request);
 		g_free(srch);
 		return;
 	}
@@ -974,7 +1251,6 @@ void gg_init(struct prpl *ret)
 	ret->protocol = PROTO_GADUGADU;
 	ret->options = 0;
 	ret->name = agg_name;
-	ret->checkbox = _("Send as message");
 	ret->list_icon = agg_list_icon;
 	ret->away_states = agg_away_states;
 	ret->actions = agg_actions;
@@ -992,7 +1268,7 @@ void gg_init(struct prpl *ret)
 	ret->get_dir = agg_get_info;
 	ret->dir_search = agg_dir_search;
 	ret->set_idle = NULL;
-	ret->change_passwd = NULL;
+	ret->change_passwd = agg_change_passwd;
 	ret->add_buddy = agg_add_buddy;
 	ret->add_buddies = agg_add_buddies;
 	ret->remove_buddy = agg_rem_buddy;
@@ -1038,3 +1314,13 @@ char *description()
 }
 
 #endif
+
+/*
+ * Local variables:
+ * c-indentation-style: k&r
+ * c-basic-offset: 8
+ * indent-tabs-mode: notnil
+ * End:
+ *
+ * vim: shiftwidth=8:
+ */
